package RFC::BaseRelay;
#$Id: BaseRelay.pm,v 1.67 2015/06/18 11:10:59 sqko Exp $


###########################################################################
#
#
#      RFC::BaseRelay
#
#
#  This class defines the behaviour of simple relays
#  To obtain a relay instance use one of the constructors like shown below:
#
#  my $id = BaseRelay->new($name,$rig);
#
#  The rig argument must be a RFC::Rig instance or a similar object honoring
#  the get_conf() member function.
#  The constructior can also be called with additional arguments:
#
#  my $id = BaseRelay->new($name,$rig,$tty,$address,$channel);
#
#  where the last 3 argmeutns specify the serial device, the module address 
#  and the channel number of the relay (as most relay modules contains more 
#  than one relay.
#
#  All relay objects have the following member functions (excluding those 
#  derived from RFC::BaseDevce and Debug.pm):
#
#  $id->set($[$][$])    Sets the status of the relay to the specified arguemnt
#                       one indicates closed relay, 0 open.
#                       If an additional argument is specified, it is assumed to be
#                       the username of the controlling user and that name is appended
#                       to the string appended to the rigs proglog.
#                       an exception to this is if a further argument is specified in 
#                       which case no proglog entry is appended! 
#
#  $id->set_noinfo($)   Similar to set() except no proglog info is appended.
#
#  If no data can be read from the device attatched to the object, the read() 
#  function returns -1 to indicate the error.
#  
####################################################################



use strict;
use warnings;
use RFC::BaseDevice;
use RFC::Cache;
our @ISA = qw(RFC::BaseDevice);

BEGIN {
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = 1.0;
    %EXPORT_TAGS = (
        'standard'=>[],
                    );
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_type'} = 'BaseRelay';
    $self->{'_mode'} = 'Automatic';
    $self->{'_setup_tags'} = ['mode','tty','address','channel','title','show_plot','use_cache'];
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'tty' => 'Serial device for communication',
	'address' => 'Device address',
	'channel' => 'Channel number on device (Note 0 based)',
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
	'mode' => 'Automatic',
	'tty' => 'ttyS0',
	'address' => 0,
	'channel' => 0,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_set_fn_names'} = ['set']; # List of function names which can trigger a notify event
    $self->{'_values_mode'} = ['Automatic'];
    $self->{'_tty'} = 'ttyS0';
    $self->{'_address'} = 0;
    $self->{'_channel'} = 0;
    bless $self,$class;
    $self->{'_section'} = 'relay_'.$self->name;
    $self->{'_title'} = $self->get_conf('title');
    return $self;
}

sub set_noinfo($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status") if ($self->debug > 2);
    $self->die($self->{'_type'}." is a base class and does not impelemnt id->set(), only the derived classes does so!");
    $self->notify('set_noinfo',$status,@_);
    return undef;
}

sub set($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status") if ($self->debug > 2);
    $self->die($self->{'_type'}." is a base class and does not impelemnt id->set(), only the derived classes does so!");
    $self->notify('set',$status,@_);
    return undef;
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Type: '.$self->type;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    push @arr,'channel: '.$self->{'_channel'};
    return join (', ',@arr);
}


package RFC::ManualRelay;
use RFC::Header;
our @ISA = qw(RFC::BaseRelay);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'_type'} = 'ManualRelay';
    $self->{'_mode'} = 'Manual';
    $self->{'_setup_tags'} = ['title','show_plot'];
    $self->{'_desc'} = {
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
    };
    $self->{'_default'} = {
	'show_plot' => 'Yes',
    };
    bless $self,$class;
    $self->{'_title'} = $self->get_conf('title');
    $self->{'_file'} = SemaforeFile->new($userdir.'rig'.$self->rig->name.'/relay_'.$self->name,'/home/celltest/Gas_rig'.$self->rig->name.'.lock');
    unless ($self->{'_file'}->exist)
    {
        ### Create file with rw for all           ###
        ### Yes I know 666 is the number of evil! ###
        $self->{'_file'}->lock_ex;
        $self->{'_file'}->writeline('0');
        chmod 0666, $self->{'_file'}->filename;
        $self->{'_file'}->unlock;
    }
    return $self;
}

sub DESTROY
{
    my $self = shift;
    $self->{'_file'}->DESTROY() if ($self->{'_file'});
    return undef;
}

sub test()
{
    return 1;
}

sub test_string()
{
    my $self = shift;
    my $val = $self->read();
    return $self->name.' is a virtual relay and thus physically controlled manually! Status for relay: '.$val;
}

sub set($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status manually!") if ($self->debug > 2);
    $self->{'_file'}->writeline($status);
    $self->notify('set',$status,@_);
    $self->proglog("Relay ".$self->name." set to $status manually",@_);
    return $status;
}

sub set_noinfo($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status manually!") if ($self->debug > 2);
    $self->{'_file'}->writeline($status);
    $self->notify('set_noinfo',$status,@_);
    return $status;
}

sub read_nocache
{
    my $self = shift;
    $self->{'_last_read'} = Time::HiRes::time;
    return $self->read(@_);
}

sub read
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $flow = 0;
    my $cached = $data->get('manual_relay',$self->name);
    if (defined($cached))
    {
        $self->writedebug("Returnming already cached value for relay ".$self->name) if ($self->debug);
        return $cached;
    }
    else
    {
        $self->writedebug("reading status from file") if ($self->debug);
        my @lines = $self->{'_file'}->readlines;
        $flow = $lines[0] if (@lines);
        $data->set('manual_relay',$self->name,$flow);
        return $flow;
    }
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Type: '.$self->type;
    push @arr,'Mode: '.$self->mode;
    return join (', ',@arr);
}



package RFC::ICPRelay;
our @ISA = qw(RFC::BaseRelay);

sub new()
{
    my $class = shift;
    my $name = shift;
    my $rig = shift;
    my $self = $class->SUPER::new($name,$rig);
    $self->{'_type'} = 'ICPRelay';
    bless $self,$class;
    foreach my $key ('tty','address','channel','title','show_plot','use_cache')
    {
	$self->{'_'.$key} = $self->get_conf($key);
	$self->{'_'.$key} = $self->get_default($key) unless ($self->{'_'.$key});
    }
    $self->{'_tty'} = shift if (@_);
    $self->{'_address'} = shift if (@_);
    $self->{'_channel'} = shift if (@_);
    $self->{'_channel'} = $self->{'_default'}{'channel'} unless ($self->{'_channel'});
    $self->{'_address'} = $self->{'_default'}{'address'} unless ($self->{'_address'});
    $self->{'_tty'} = $self->{'_default'}{'tty'} unless ($self->{'_tty'});
    return $self;
}

sub test()
{
    my $self = shift;
    my $res = undef;
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    $res = SocketClient::serial_client($self->{'_tty'},'multiplex','read',$self->{'_address'});
    return 0 unless (defined($res) && $res =~ /\d+/ && $res !~ /error/i);
    my @dat = split(//,$res);
    return 0 unless (@dat > $self->{'_channel'});
    return 1;
}

sub test_string()
{
    my $self = shift;
    my $retval = undef;
    eval {
	$retval = 'Status for relay '.$self->name.': '.$self->read(@_);
    };
    if ($@)
    {
	return 'Error: could not get status for relay '.$self->name.', error: '.$@;
    }
    return $retval;
}


sub set($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status") if ($self->debug > 2);
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'RELAY',$self->{'_address'},$self->{'_channel'},$status);
    $self->writedebug("Response for relay ".$self->name.": $res") if ($self->debug > 3);
    $self->notify('set',$status,@_);
    $self->proglog("Relay ".$self->name." set to $status",@_);
    return $res;
}

sub set_noinfo($)
{
    my $self = shift;
    my $status = shift;
    $self->writedebug("Setting status for relay ".$self->name." to $status") if ($self->debug > 2);
    my $res = SocketClient::serial_client_raw($self->{'_tty'},'RELAY',$self->{'_address'},$self->{'_channel'},$status);
    $self->writedebug("Response for relay ".$self->name.": $res") if ($self->debug > 3);
    $self->notify('set_noinfo',$status);
    return $res;
}

sub read_nocache()
{
    my $self = shift;
    my $data = undef;
    if (@_)
    {
        $data = $_[0];
    }
    else
    {
	$data = RFC::Cache->new();
	$data->debug($self->debug()) if ($self->debug);
    }
    $self->check_cache($data);
    my $addr = $self->{'_address'};
    my $tty = $self->{'_tty'};
    my $cached = $data->get($tty,'icp',$addr,$self->{'_channel'});
    if ($cached)
    {
	$self->writedebug("Using existing data from relay on tty $tty address $addr") if ($self->debug > 3);
	$self->{'_last_read'} = Time::HiRes::time;
	return $cached;
    }
    else
    {
	$self->writedebug("Reading data from relay on tty $tty address $addr") if ($self->debug > 3);
	my $res = SocketClient::serial_client($self->{'_tty'},'multiplex','read',$self->{'_address'});
	unless (defined($res))
	{
	    ### Retry once if no data ##
	    select(undef,undef,undef,0.1);
	    $res = SocketClient::serial_client($self->{'_tty'},'multiplex','read',$self->{'_address'});
	}
	if (defined($res) && $res !~ /error/i)
	{
	    chomp $res;
	    my @dat = split(//,$res);
	    for (my $x = 0; $x < @dat; $x++)
	    {
		$data->set($tty,'icp',$addr,$x,$dat[$x]);
	    }
	    $self->writedebug("Status for relay ".$self->name." was ".$dat[$self->{'_channel'}]) if ($self->debug > 2);
	    my $str = time .' '.$self->name.' '.$dat[$self->{'_channel'}];
	    $self->rig->kanaldata($str)  unless ($self->readstring_ignore);
	    $self->set_cache($dat[$self->{'_channel'}]);
	    $self->{'_last_read'} = Time::HiRes::time;
	    return $dat[$self->{'_channel'}];
	}
	else
	{
	    $self->writedebug("Warning: No data returned from relay device ".$self->name,1);
	    $data->set($tty,'icp',$addr,$self->{'_channel'},-1);
	    return -1;
	}
    }
}



package RFC::ICPDI;
our @ISA = qw(RFC::ICPRelay);

sub new()
{
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    my $section = shift;
    $self->{'_type'} = 'ICPDI';
    $self->{'_readonly'} = 1;
    $self->{'_mode'} = 'Readonly';
    $self->{'_setup_tags'} = ['mode','tty','address','channel','title','show_plot','use_cache'];
    $self->{'_desc'} = {
	'mode' => 'Device control mode',
	'tty' => 'Serial device for communication',
	'address' => 'Device address',
	'channel' => 'Channel number on device (Note 0 based)',
	'title' => 'Optional title',
        'show_plot'=>'Determines if the current device data is to be shown in the daily data plots',
        'use_cache'=>'Determines if persisten caching of read values is allowed',
    };
    $self->{'_default'} = {
	'mode' => 'Readonly',
	'tty' => 'ttyS0',
	'address' => 0,
	'channel' => 0,
        'show_plot'=>'Yes',
        'use_cache'=>'No',
    };
    $self->{'_values_mode'} = ['Readonly'];
    bless $self,$class;
    return $self;
}

sub set($)
{
    my $self = shift;
    $self->die("Digital input ".$self->name." can not be used for output!");
    return undef;
}

sub set_noinfo($)
{
    my $self = shift;
    $self->die("Digital input ".$self->name." can not be used for output!");
    return undef;
}

sub print_config()
{
    my $self = shift;
    my @arr = ();
    push @arr,'Name: '.$self->name;
    push @arr,'Type: '.$self->type;
    push @arr,'TTY: '.$self->{'_tty'};
    push @arr,'Address: '.$self->{'_address'};
    push @arr,'channel: '.$self->{'_channel'};
    return join (', ',@arr);
}

return 1;
